;=============================================================
; RS232 self-powered projector status LED
;  - Protocol: Hitachi (CP-X450 and compatible models),
;              polls status twice per second.
;  - Bicolor LED states:
;     * Steady red:   Projector is turned off
;     * Steady green: Projector is turned on
;     * Blinking red: Projector is cooling down
;
; MCU: PIC12F615
;  (may be rewritten for PIC10F with some work on the timer)
;
; Pin assignments:
;  - RED_LED_PIN:   +5V output - power for red LED
;  - GREEN_LED_PIN: +5V output - power for green LED
;  - TX_PIN:        Logic output - RS232 data out (0-5V)
;  - RX_PIN:        Logic input  - RS232 data in  (0-5V)
;
; Peripherals used:
;  - Timer2:        500ms timer for polling interval
; 
; Protocol specification:
;  - Command to retrieve power status:
;      BE EF 03 06 00 19 D3 02 00 00 60 00 00
;  - Expected responses:
;      1D 00 = Off
;      1D 01 = On
;      1D 02 = Cooling down
;=============================================================


	list p=12f615
	radix dec
#include "p12f615.inc"

    __config _INTRC_OSC_NOCLKOUT & _WDT_OFF & _PWRTE_OFF & _MCLRE_OFF & _CP_OFF & _IOSCFS4 & _BOR_OFF

	org 0x0000       ;Reset vector
	goto main

	org 0x0004       ;Interrupt vector
    retfie

;=============================================================
; Constants
;=============================================================
RED_LED_PIN   equ 1
GREEN_LED_PIN equ 0
TX_PIN        equ 4
RX_PIN        equ 5
GPPU          equ 7

TRIS_MASK     equ 0x2C   ;Pins GP0,GP1 and GP4 are outputs

;=============================================================
; Global variables
;=============================================================
response_expected equ 0x7F	;Filters out unsolicited received bytes
blink_state       equ 0x7E	;Toggle flag for blinking LED
counter           equ 0x7D	;Timer cycle counter

;=============================================================
; Local variables
;=============================================================
char_to_send  equ 0x40 ;RS-232 transmit
received_char equ 0x40 ;RS-232 receive
delay_low     equ 0x40 ;Delay loops (ms)

bit_count     equ 0x41 ;RS-232 receive/transmit
delay_high    equ 0x41 ;Delay loops (ms)

delay_count   equ 0x42 ;Delay loops (us)

response_char equ 0x43 ;Response character parsing


	org 0x0005

main:

;=============================================================
; I. Startup configuration
;     - Initialize I/O pins:
;         GP0: Output, low
;         GP1: Output, low
;     - Initialize Timer2
;=============================================================

	clrf  GPIO
    clrf  response_expected
    clrf  blink_state
    movlw 25
    movwf counter

	bsf   STATUS, RP0    ; Select bank 1

    bcf   OPTION_REG, GPPU  ; Enable pull-up resistor control
    ;bsf   WPU, RX_PIN       ; Enable weak pull-up on the RX input pin
    
    clrf  ANSEL          ; Set all pins to digital I/O
    movlw TRIS_MASK
    movwf TRISIO         ; Set output pins to output mode

    movlw 125            ;Timer2 period = 125 (20ms = 500ms/25)
    movwf PR2
    
	bcf   STATUS, RP0    ; Select bank 0

    call delay_100ms

    call transmit_command

    clrf  TMR2           ;Set Timer2 prescaler = 16, postscaler = 10
    bcf   PIR1, TMR2IF
    movlw 0x4F           ;0 1001 1 11
    movwf T2CON

;=============================================================
; II. Main loop
;     - Check for periodic timer
;     - Check for RS-232 start bit
;=============================================================
main_loop:
    btfsc PIR1, TMR2IF   ;Manually check for interrupt flag
    call handle_timer

    btfss GPIO, RX_PIN   ;
    goto rx_active

    goto main_loop

;=============================================================
; RS-232 start bit detected:
;  - Receive the byte
;  - If expected, act on the byte
;=============================================================
rx_active:
    call receive_byte
    movwf response_char
    btfss response_expected, 0   ; Skip if no response is expected
    goto main_loop

    clrf response_expected       ; Clear the flag now that we have a response

    movfw response_char          ; Set the flag to continue reading if the response
    xorlw 0x1D                   ;  matches the first byte (0x1D).
    btfsc STATUS, Z
    bsf  response_expected, 0

    movfw response_char          ; Response 1: Power state is on
    xorlw 0x01                   ;   - Light the green LED
    btfsc STATUS, Z
    call show_green_led

    movfw response_char          ; Response 0: Power state is off
    xorlw 0x00                   ;   - Light the red LED
    btfsc STATUS, Z
    call show_red_led

    movfw response_char          ; Response 2: Power state is intermediate
    xorlw 0x02                   ;  (Lamp is cooling down)
    btfss STATUS, Z              ;  - Blink the red LED
    goto main_loop

    call unshow_leds

    movlw 0x01
    xorwf blink_state, F
    btfsc STATUS, Z
    call show_red_led
    
    goto main_loop

;=============================================================
; 20ms timer event detected:
;  - Count up to 500ms
;=============================================================
handle_timer:
    bcf    PIR1, TMR2IF
    decfsz counter,F
    return

;=============================================================
; 500ms timer event detected:
;  - Reset the counter
;  - Transmit the command to retrieve the power status
;=============================================================
    movlw 25
    movwf counter

transmit_command:
    movlw 0xBE
    call transmit_byte
    movlw 0xEF
    call transmit_byte
    movlw 0x03
    call transmit_byte
    movlw 0x06
    call transmit_byte
    movlw 0x00
    call transmit_byte
    movlw 0x19
    call transmit_byte
    movlw 0xD3
    call transmit_byte
    movlw 0x02
    call transmit_byte
    movlw 0x00
    call transmit_byte
    movlw 0x00
    call transmit_byte
    movlw 0x60
    call transmit_byte
    movlw 0x00
    call transmit_byte
    movlw 0x00
    call transmit_byte
    bsf response_expected, 0
    bcf   PIR1, TMR2IF
    return

;=============================================================
; RS232 functions
;  19200bps = 52us bit-to-bit
;=============================================================
transmit_byte:  ;w = byte to transmit
    movwf char_to_send       ;Store byte in temporary memory
    movlw 0x8
    movwf bit_count    ;Set count to 8

    call delay_50_cycles

    bsf GPIO, TX_PIN     ;Send the start bit (set)
    call delay_42_cycles        ;1 + 48 + 3 = 52
    nop
    nop
    nop
    nop

transmit_bit:
    btfss char_to_send, 0   ;3 cycles whether branch is taken or not
    goto transmit_0bit
    nop
    
transmit_1bit:
    bcf GPIO, TX_PIN        ;1 cycle (set)
    goto transmit_bit_done  ;2 cycles
    
transmit_0bit:
    bsf GPIO, TX_PIN        ;1 cycle (set)
    nop                     ;2 cycles
    nop

transmit_bit_done:
   call delay_42_cycles     ;42 cycles
   rrf  char_to_send, F     ;4 cycles if branch taken, 3 cycles if not
   decfsz bit_count, F
   goto transmit_bit                    ;1 + 2 + 42 + 4 + 3 = 52

   nop  ;1 + 2 + 42 + 3 + 4 = 52
   nop
   nop
   nop
   bcf GPIO, TX_PIN

   return
   

receive_byte:           ; 3 + 2 cycles for call
   clrf received_char            ; 3 cycles set up
   movlw 0x8
   movwf bit_count
   call delay_18_cycles  ;26 cycles from call to bit-test
   btfsc GPIO, RX_PIN    ;2 cycles (set)
   goto receive_error
   nop                   ;4 cycles
   nop
   nop
   nop
   nop

receive_bit:
   nop
   nop
   rrf  received_char, F          ;2 cycles (including nop)
   call delay_42_cycles  ;52 cycles from bit-test to bit_test
   btfsc GPIO, RX_PIN    ;3 cycles (set)
   goto receive_1bit
   nop

receive_0bit:
   bcf received_char, 7           ;1 cycle
   decfsz bit_count,F         ;3 cycles if not done, 4 if done
   goto receive_bit
   goto receive_done

receive_1bit:
   bsf received_char, 7           ;1 cycle
   decfsz bit_count,F         ;3 cycles if not done, 4 if done
   goto receive_bit
   nop
   nop

receive_done:
   nop
   nop
   call delay_42_cycles  ;52 cycles from last bit-test to stopbit-test
   btfss GPIO, RX_PIN
   goto receive_error

   movfw received_char
   return

receive_error:
   clrw
   return
   


;=============================================================
; Support functions
;=============================================================

show_red_led:
    bcf   GPIO, GREEN_LED_PIN
    bsf   GPIO, RED_LED_PIN
    return

show_green_led:
    bcf   GPIO, RED_LED_PIN
    bsf   GPIO, GREEN_LED_PIN
    return

unshow_leds:
    bcf   GPIO, RED_LED_PIN
    bcf   GPIO, GREEN_LED_PIN
    return

delay_18_cycles:            ;2 cycles for call
    movlw 0x04              ;2 cycles for setup
    movwf delay_count
delay_18_loop:              ;3 cycles per iteration (minus 1)
    decfsz delay_count, F
    goto delay_18_loop
    nop
    return                  ;3 cycles for return

delay_42_cycles:            ;2 cycles for call
    movlw 0x0C              ;2 cycles for setup
    movwf delay_count
delay_42_loop:              ;3 cycles per iteration (minus 1)
    decfsz delay_count, F
    goto delay_42_loop
    nop
    return                  ;3 cycles for return

delay_50_cycles:            ;2 cycles for call
    movlw 0x0F              ;2 cycles for setup
    movwf delay_count
delay_50_loop:              ;3 cycles per iteration (minus 1)
    decfsz delay_count, F
    goto delay_50_loop
    return                  ;2 cycles for return

delay_5ms:
    movlw 0x8B
    movwf delay_low
    movlw 0x06
    movwf delay_high
    goto delay_loop

delay_100ms:
    movlw 0xDD
	movwf delay_low
	movlw 0x82
	movwf delay_high
delay_loop:
	decfsz delay_low, F
	goto delay_loop
    clrwdt
	decfsz delay_high, F
	goto delay_loop
	return
	end